#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include <algorithm>

#include "ihx.h"

#define IHX_MAX_LINE_LEN	(256*2+12)
#define BINREAD_BLOCK_SIZE	1024

using namespace ihx;

void Buffer::AddElement(const BuffElem &be) {
	std::vector<BuffElem>::iterator i, e;

	// izbrise sve koje u potpunosti prekriva
ponovo_trazi:
	for (i = begin(); i != end(); ++i) {
		if (be.Overrides(*i)) {
			erase(i);
			goto ponovo_trazi;
		}
	}

	// doda novi komad
	push_back(be);
	e = end() - 1;

	// pokusava ga mergeati
	for (i = begin(); i != end(); ++i) {
		if (e != i && (e->Covers(*i) || e->Touches(*i))) {
			unsigned int pos;
			i->Merge(*e);
			if (e > i) {
				pos = i - begin();
			} else {
				pos = i - begin() - 1;
			}
			erase(e);
			e = i = begin() + pos;
		}
	}
}

Buffer::Error Buffer::ReadIHX(const char *path) {
	unsigned char bytes, type, chk, chk2;
	unsigned char ml[IHX_MAX_LINE_LEN / 2];
	char line[IHX_MAX_LINE_LEN];
	unsigned short int address, addr_hi = 0, addr_seg = 0;
	FILE *fi;

	fi = fopen(path, "rt");
	if (!fi)
		return ERR_OPEN;

	while (fgets(line, IHX_MAX_LINE_LEN, fi)) {
		if (*line != ':') {
			fclose(fi);
			return ERR_FILEFORMAT;
		}
		sscanf(line + 1, "%2hhX%4hX%2hhX", &bytes, &address, &type);
		chk = bytes + (address & 0xFF) + (address >> 8) + type;
		for (int i = 0; i < bytes; ++i) {
			sscanf(line + 9 + i * 2, "%2hhX", &ml[i]);
			chk += ml[i];
		}
		sscanf(line + 9 + bytes * 2, "%2hhX", &chk2);
		if ((unsigned char)(~chk + 1) != chk2) {
			fclose(fi);
			return ERR_CHECKSUM;
		}

		switch (type) {
			case 0x00:	// data
				AddElement(BuffElem((address | (addr_hi << 16)) + addr_seg * 16, ml, bytes));
				break;
			case 0x01:	// eof
				goto endwhile;
			case 0x02:	// 16 bit segment
				if (address != 0 || bytes != 2)
					goto rec_err;
				addr_seg = (ml[0] << 8) | ml[1];
				break;
			case 0x04:	// address high 16 bits
				if (address != 0 || bytes != 2)
					goto rec_err;
				addr_hi = (ml[0] << 8) | ml[1];
				break;
			case 0x03:
			case 0x05:	// start address (for CPU)
				if (address != 0 || bytes != 4)
					goto rec_err;
				eip = (ml[0] << 24) | (ml[1] << 16) | (ml[2] << 8) | ml[3];
				break;
			default:
rec_err:
				fclose(fi);
				return ERR_UNKNOWNREC;
		}
	}
endwhile:

	fclose(fi);
	return ERR_OK;
}

Buffer::Error Buffer::ReadBIN(const char *fname, unsigned int offset) {
	FILE *fi;
	unsigned char *buf = NULL;
	size_t sz = 0, l;

	if (!(fi = fopen(fname, "rb")))
		return ERR_OPEN;

	do {
		buf = (unsigned char*)realloc(buf, sz + BINREAD_BLOCK_SIZE);
		l = fread(buf + sz, 1, BINREAD_BLOCK_SIZE, fi);
		sz += l;
	} while (l);

	fclose(fi);

	AddElement(BuffElem(offset, buf, sz));
	free(buf);

	return ERR_OK;
}

Buffer::Error Buffer::WriteBIN(const char *fname) const {
	FILE *fo;

	fo = fopen(fname, "wb");
	if (!fo)
		return ERR_OPEN;

	for (const_iterator i = begin(); i != end(); ++i) {
		fseek(fo, i->address, SEEK_SET);
		fwrite(i->GetBuffer(), i->GetSize(), 1, fo);
	}

	fclose(fo);
	return ERR_OK;
}

void * Buffer::GetBINBlock(unsigned int size, unsigned int offset, unsigned char fill) const {
	unsigned char *m = (unsigned char *)malloc(size);

	memset(m, fill, size);
	for (const_iterator i = begin(); i != end(); ++i) {
		unsigned int bla = i->address + i->GetSize() - 1;	// block last address

		if (i->address < offset && bla >= offset) {	// jedino kada i->address < offset
			unsigned int fofs = offset - i->address;	// fizički offset
			memcpy(m, i->GetBuffer() + fofs, std::min(i->GetSize() - fofs, size));
		} else if (i->address >= offset) {		// inače uvijek mora i->address >= offset
			unsigned int vma = i->address - offset;		// adresa na koju sprema

			if (vma < size)
				memcpy(m + vma, i->GetBuffer(), std::min(i->GetSize(), size - vma));
		}
	}

	return m;
}

void Buffer::FreeBINBlock(void *m) {
	free(m);
}

unsigned int Buffer::CalcSize() const {
	unsigned int sz = 0;
	for (const_iterator i = begin(); i != end(); ++i)
		sz += i->GetSize();
	return sz;
}

unsigned int Buffer::LastAddress() const {
	unsigned int la = 0;
	for (const_iterator i = begin(); i != end(); ++i)
		if (la < i->address + i->GetSize())
			la = i->address + i->GetSize();
	return la - 1;
}

const char * Buffer::GetErrorString(Error e) {
	static const char *greske[] = {
		"No error",
		"Can't read/write file",
		"Invalid file format",
		"Invalid checksum",
		"Unknown record type"
	};

	if (e < 0 || e > (int)(sizeof(greske) / sizeof(char*)))
		return "Invalid error number";
	return greske[e];
}
